#include <pspkernel.h>
#include <pspaudio.h>

#include <stdlib.h>
#include <string.h>

#include "pspaudio.h"

#define NUM_AUDIO_BUFFERS 16
#define AUDIO_BUFFER_SIZE 1024	//	psp buffers must be multiples of 64


const int num_audio_buffs = NUM_AUDIO_BUFFERS;
const int audio_buff_size = AUDIO_BUFFER_SIZE;

//	the buffer
short audio_buff[2 * AUDIO_BUFFER_SIZE * NUM_AUDIO_BUFFERS];

//	1=44100, 2=22050, 4=11025
int sample_rate_div = 1;

int audio_handle = -1;
int audio_sema = -1;
int audio_thread = 0;

int playing_buffer = 0;
int write_offset = 0;

int audio_active = 1;

static int buffered_samples = 0 ;

int AudioThreadFunc(int args, void* argp)
{
  int cacheflushasked=0;

	while (audio_active)
	{
 	  if ( cacheflushasked)
 	  {
 	    sceKernelDcacheWritebackAll();
 	    cacheflushasked = 0;
 	  }
		
		sceAudioOutputPannedBlocking(audio_handle, PSP_AUDIO_VOLUME_MAX, PSP_AUDIO_VOLUME_MAX, &audio_buff[playing_buffer * 2 * AUDIO_BUFFER_SIZE]);

    // -- Check if buffered samples are more than 3 frames ahead
    if ( buffered_samples >= 4*AUDIO_BUFFER_SIZE )
    {
      buffered_samples -= 2*AUDIO_BUFFER_SIZE; // If yes, then skip one frame
      
      playing_buffer+=2;
      
  		if (playing_buffer >= NUM_AUDIO_BUFFERS)
  			playing_buffer -= NUM_AUDIO_BUFFERS;
  			
  		cacheflushasked=1;
    }
    else if ( buffered_samples >= AUDIO_BUFFER_SIZE ) // If buffer is one or two frame ahead, then play next buffer
    {
      buffered_samples -= AUDIO_BUFFER_SIZE;
      
      playing_buffer++;
      
  		if (playing_buffer >= NUM_AUDIO_BUFFERS)
  			playing_buffer -= NUM_AUDIO_BUFFERS;
  			
  	  cacheflushasked=1;
    }
    // -- If not enough samples, then play again same buffer

	}
	sceKernelExitThread(0);
	return 0;
}

void pspaudio_init(void)
{
	pspaudio_clear();

	if (audio_handle == -1)
		audio_handle = sceAudioChReserve(-1, AUDIO_BUFFER_SIZE, 0);

    audio_thread = sceKernelCreateThread("audio_thread", (void*)AudioThreadFunc, 0x12, 0x10000, 0, 0);
    if (audio_thread >= 0)
		sceKernelStartThread(audio_thread, 0, 0);

	if (audio_sema == -1)
		audio_sema = sceKernelCreateSema("audio_buffer_1", 0, 1, 1, 0);	// breaks without this, why?..
}

void pspaudio_clear(void)
{
	int buff;
	for (buff = 0; buff < NUM_AUDIO_BUFFERS; buff++)
		memset(&audio_buff[buff * 2 * AUDIO_BUFFER_SIZE], 0, 2 * AUDIO_BUFFER_SIZE * sizeof(short) );
}

void pspaudio_write(short* buffer, int num_samples)
{
	//	could be faster...
	
	buffered_samples += num_samples;
	
	int i;
	if (sample_rate_div == 1) // 44100
	{
		for (i = 0; i < num_samples; i++)
		{
			short s = buffer[i];
			//	l+r
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;

			if (write_offset == 2 * AUDIO_BUFFER_SIZE * NUM_AUDIO_BUFFERS)
				write_offset = 0;
		}
	}
	else if (sample_rate_div == 2) // 22050
	{
		for (i = 0; i < num_samples; i++)
		{
			short s = buffer[i];
			//	upsample, no interp
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;

			if (write_offset == 2 * AUDIO_BUFFER_SIZE * NUM_AUDIO_BUFFERS)
				write_offset = 0;
		}
	}
	else // else 11025
	{
		for (i = 0; i < num_samples; i++)
		{
			short s = buffer[i];
			//	upsample, no interp
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;
			audio_buff[write_offset++] = s;

			if (write_offset == 2 * AUDIO_BUFFER_SIZE * NUM_AUDIO_BUFFERS)
				write_offset = 0;
		}
	}

	return;
}

void pspaudio_shutdown(void)
{
	//	signal thread to end
	audio_active = 0;
	audio_thread = 0;
}
