/*
 * vsync.c - End-of-frame handling for MS-DOS.
 *
 * Written by
 *  Ettore Perazzoli <ettore@comm2000.it>
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */

//void pspaudiologframe(float fps);

#include "vice.h"

#include "clkguard.h"
#include "cmdline.h"
#include "drive.h"
#include "machine.h"
#include "maincpu.h"
#include "resources.h"
#include "videoarch.h"
#include "sound.h"
#include "ui.h"
#include "types.h"
#include "vsync.h"
#include "vkeyboard.h"
#include "kbd.h"
#include "joy.h"
#include "menu.h"
#include "fileexp.h"
#include "infowin.h"
#include "pad.h"
#include "osdmsg.h"
#include "joy.h"
#include "kbdbuf.h"

#include "sys/time.h"

#include "PSPSpecific.h"
#include "pspaudio.h"

extern int in_gameloop;
void DisplayDebugMessages();

extern unsigned char *ps2_videobuf;

/* Maximum number of frames we can skip consecutively when adjusting the
   refresh rate dynamically.  */
#define MAX_SKIPPED_FRAMES	10

/* ------------------------------------------------------------------------- */

/* Vsync-related resources.  */
static const resource_t resources[] = {
    { NULL }
};

int vsync_resources_init(void)
{
    return resources_register(resources);
}

/* ------------------------------------------------------------------------- */

/* Vsync-related command-line options.  */
static const cmdline_option_t cmdline_options[] = {
    { "-speed", SET_RESOURCE, 1, NULL, NULL, "Speed", NULL,
      "<percent>", "Limit emulation speed to specified value" },
    { "-refresh", SET_RESOURCE, 1, NULL, NULL, "RefreshRate", NULL,
      "<value>", "Update every <value> frames (`0' for automatic)" },
    { "-warp", SET_RESOURCE, 0, NULL, NULL, "WarpMode", (resource_value_t)1,
      NULL, "Enable warp mode" },
    { "+warp", SET_RESOURCE, 0, NULL, NULL, "WarpMode", (resource_value_t)0,
      NULL, "Disable warp mode" },
    { NULL }
};

int vsync_cmdline_options_init(void)
{
    return cmdline_register_options(cmdline_options);
}

/* ------------------------------------------------------------------------- */

/* Number of frames per second on the real machine.  */
static float refresh_frequency;

/* Number of clock cycles per seconds on the real machine.  */
static long cycles_per_sec;

/* Function to call at the end of every screen frame.  */
static void (*vsync_hook)(void);

/* ------------------------------------------------------------------------- */

static int timer_speed = -1;

//int ps2_framecount = 0;

/* Speed evaluation. */

static CLOCK speed_eval_prev_clk;

static int speed_eval_suspended = 1;

void vsync_suspend_speed_eval(void)
{
    speed_eval_suspended = 1;
}

void vsync_sync_reset(void)
{
    speed_eval_suspended = 1;
}

static void calc_avg_performance(int num_frames)
{
}

float vsync_get_avg_frame_rate(void)
{
  return 0;
}

float vsync_get_avg_speed_index(void)
{
  return 0;
}

/* ------------------------------------------------------------------------- */

/* This prevents the clock counters from overflowing.  */
static void clk_overflow_callback(CLOCK amount, void *data)
{
    speed_eval_prev_clk -= amount;
}

//int disp_w = 512;
//int disp_h = 512; 
int disp_w = 480;
int disp_h = 272; 

/* ------------------------------------------------------------------------- */

int is_paused = 0;

//int previous_vcont;
//int current_vcont;
//float fps;

float lastvalue;
float elaspsed_time;


int wait_1ms()
{
  int i;
  int a;
  float b;
  a=0;
  b=0.0f;
  for (i=0; i<100000; i++)
  {
    a=i;
    b=b*a;
    b=b/a;
  }
  return b;
}


//extern int lastkeyfound;
  
int vsync_do_vsync(struct video_canvas_s* c, int been_skipped)
{

  // -- Get Time
	/*
	struct timeval tval;
	float curr;
	gettimeofday(&tval,0);
	curr = tval.tv_sec + (((float)tval.tv_usec) / 1000000);  // Get absolute time in seconds
  elaspsed_time = curr - lastvalue;
  lastvalue = curr;
  */

  // -- Pause thread to have exact timing (usec = micro second, 10-6s)
  //if ( elaspsed_time < 0.020f )
  //  sceKernelSleepThread( (0.020f - elaspsed_time) * 1000000.0f );

  /*
  int sleeptime;
  sleeptime = (int) ((0.020f - elaspsed_time) * 1000000.0f);
  
  if ( sleeptime < 0 )
    sleeptime = 1;
    
  if ( sleeptime > 1000000 )
    sleeptime = 1000000;


  //wait_5ms();
  int i;
  int a;
  float b;
  a=0;
  b=0.0f;
  for (i=0; i<900000; i++)
  {
    a=i;
    b=b*a;
    b=b/a;
  }  
  */

  //sceKernelDelayThread( 5000 );
  
  //if ( (elaspsed_time > 0.0f ) && (elaspsed_time < 0.020f) )
  //{
    //sceKernelDelayThread( sleeptime );
    //sceKernelDelayThread(10000);
  //}
  
  //sceKernelDelayThread(10000);

  //printf("vsync_do_vsync");

  vsync_hook();

  //ps2_framecount++;

  do
  {
    //in_gameloop=1;

    PSPDisplayImage_Reset();

    PSPStartDisplay();

    kbd_update();
    pad_update();

    PSPDisplayVideoBuffer();

    //if ( (current_vcont - previous_vcont) != 0 )
    //  fps = 60.0f/(current_vcont - previous_vcont); 

    //char message[256];
    //sprintf(message, "Time : %f", elaspsed_time);
    //PSPDisplayText(480-(20*8),  8, 0x0ffffff, message); 

    //char message[32];
    //sprintf(message, "%d", lastkeyfound);
    //PSPDisplayText(480-(20*8),  8, 0x0ffffff, message); 
    

	  /*
    sprintf(message, "X : %d", buttons.ljoy_h );
    PSPDisplayText(480-(10*8),  16, 0x0ffffff, message); 
    sprintf(message, "Y : %d", buttons.ljoy_v );
    PSPDisplayText(480-(10*8),  24, 0x0ffffff, message); 
	  */
	
	/*
    sprintf(message, "X : %02x", paddata1);
    PSPDisplayText(480-(10*8),  16, 0x0ffffff, message); 
	*/


    // Draw message "Pause"
    if ( (is_paused) && (menu_active==0) )
    {
      draw_msg("PAUSED");
      //pspaudio_clear();
    }

    if (aboutwin_active)
    {
      aboutwin_frame();
    } 
    else if (infowin_active)
    {
      infowin_frame();
    }
    else if (file_explorer_active)
    {
      file_explorer_frame();
    }
    else if (!menu_update())
    {
      vkeyboard_frame();
      infomsg_frame();
      
      // Do not write single { and }, it make crashs !!
      
      //if (new_pad1 || new_rel1) // CK : Force update
      //{
        joy_setabs();
      //}
      //if (new_pad2 || new_rel2) // CK : Force update
      //{
        joy2_setabs();
      //}
      
      
      if ( !vkeyboard_IsActive() )
      {
        // Test special buttons
        // Square = Space
        // Circle = Enter
        
        // 43 = H
        // 44 = K
        // 45 = :
        
         if (new_pad & PAD_SQUARE)
         {
           kbd_press_button(/*K_SPACE+1*/ 32);
         }
         else if ( new_rel & PAD_SQUARE )
         {
           kbd_rel_button(/*K_SPACE+1*/ 32);
         }
         
         // 40 = Nothing
         if (new_pad & PAD_CIRCLE)
         {
          kbd_press_button(/*K_RETURN+1*/ 15);
         }
         else if ( new_rel & PAD_CIRCLE )
         {
           kbd_rel_button(/*K_RETURN+1*/ 15);
         }
      }      
      
    }

    //DisplayDebugMessages(); // CK
    
    PSPEndDisplay();

    //previous_vcont = current_vcont;
		//current_vcont = PSPGetVBLCount();
		
		PSPWaitVBLAndSwap();

    //pspaudiologframe(fps);		
    //pspaudiologframe(0);	
    
    if (is_paused)
      pspaudio_clear();	

  } while (is_paused);

  sound_flush(100);

  kbd_buf_flush();

  // Wait here a bit, to have 50fps...
  //lastvalue=curr;

  return 0;
}

/* -------------------------------------------------------------------------- */

void vsync_set_machine_parameter(float refresh_rate, long cycles)
{
    refresh_frequency = refresh_rate;
    cycles_per_sec = cycles;
}

void vsync_init(void (*hook)(void))
{
  vsync_hook = hook;
}

int vsync_disable_timer(void)
{
    /* FIXME: Find a more generic solution.  */
    if (machine_class == VICE_MACHINE_CBM2)
        timer_speed = 0;
    return 0;
}

void vsyncarch_presync()
{

}


void vsyncarch_display_speed(float speed, float frame_rate, int warp_enabled) {

}

unsigned long vsyncarch_gettime(void) {
  return 0;
}
